/* File: CSFCLIP
/* File: CSFCLIP
** Description:
**   Clips multiple individual sprites from larger PCX format files.
** Copyright:
**   Copyright 1994, David G. Roberts
** Syntax:
**   CSFCLIP image[.pcx] [output-format [border-color]]
**   Where output-format equals:
**     pcx
**     planar
**     linear
**   Where border-color equals:
**     an integer from 0 to 255
*/

#include <stdlib.h>

#include "wait.h"
#include "cgdk.h"
#include "clip.h"


/*
    Function: GetPixel
    Description:
        Gets a pixel value from a bitmap at the specified coordinates.
*/
static int GetPixel(void *bitmap, int x, int y, int sizeX)
{
    unsigned char  *Data;
    unsigned Offset;

    Data = (unsigned char  *) bitmap;
    Offset = y * sizeX + x;

    return Data[Offset];
}

/*
        Function: SetPixel
    Description:
        Sets a pixel value of a bitmap at the specified coordinates.
*/
static void SetPixel(void *bitmap, int x, int y, int Color, int sizeX)
{
    unsigned char  *Data;
    unsigned Offset;

    Data = (unsigned char  *) bitmap;
    Offset = y * sizeX + x;

    Data[Offset] = Color;
}

/*
        Function: FindSprite
    Description:
        Searches the source image bitmap until it finds a color
        matching the border color and then locates the corners
        of the bounding rectangle.
*/
TCLIP_listNode *CLIP_findSprite(void *Source, int Border, int sizeX, int sizeY)
{
    unsigned Index;
    unsigned char  * Data;
    unsigned DataLen;
    int Top;
    int Left;
    int Bottom;
    int Bottom1;
    int Bottom2;
    int Right;
    int Right1;
    int Right2;
    TCLIP_listNode * Node, *Result;

  do
  {
    Data = (unsigned char  *) Source;
    DataLen = sizeX * sizeY;

    /* try to find a pixel with the border color */
    Index = 0;
    while (Index < DataLen && Data[Index] != Border) {
        Index++;
    }

    if (Index == DataLen) {
        /* we didn't find a pixel of the bounding box color */
        return NULL;
    }

    /* convert Index to x,y location of pixel */
        Top = Index / sizeX; // Source->Width;
    Left = Index % sizeX; // Source->Width;
    Bottom = Bottom1 = Bottom2 = Top;
    Right = Right1 = Right2 = Left;
    /* blacken initially found pixel */
    SetPixel(Source, Left, Top, ~Border, sizeX);

    /* the first pixel found is always the upper right (if the */
    /* border is really a rectangle), so now look for the lower */
    /* right corner */

    /* move to right from upper left */
    while (GetPixel(Source, Right1 + 1, Top, sizeX) == Border) {
        Right1++;
        SetPixel(Source, Right1, Top, ~Border, sizeX);
    }

    /* move down from upper left */
    while (GetPixel(Source, Left, Bottom1 + 1, sizeX) == Border) {
        Bottom1++;
        SetPixel(Source, Left, Bottom1, ~Border, sizeX);
    }

    /* move to right from lower left */
    while (GetPixel(Source, Right2 + 1, Bottom1, sizeX) == Border) {
        Right2++;
        SetPixel(Source, Right2, Bottom1, ~Border, sizeX);
    }

    /* move down from upper right */
    while (GetPixel(Source, Right1, Bottom2 + 1, sizeX) == Border) {
        Bottom2++;
        SetPixel(Source, Right1, Bottom2, ~Border, sizeX);
    }

    if (Right1 != Right2) {
//        fprintf(stderr, "WARNING: Found malformed sprite.\n");
    }

    if (Bottom1 != (Bottom2 + 1)) {
//        fprintf(stderr, "WARNING: Found malformed sprite.\n");
    }

    Right = _max(Right1, Right2);
    Bottom = _max(Bottom1, Bottom2);

    Node = (TCLIP_listNode *) _alloc(sizeof(TCLIP_listNode));
    if (Node == NULL) {
      return NULL;
//      CGDK_exit ("CGDK fatal error : Out of memory while searching for a sprite (CLIP_findSprite).\n\n");
    }

    Node->Rect.Top              = Top;
    Node->Rect.Left             = Left;
    Node->Rect.Bottom   = Bottom;
    Node->Rect.Right    = Right;
  }
  while ((Left == Right) || (Top == Bottom));

  return Node;
}

/*
        Function: FindAllSprites
    Description:
        Finds all the sprites in the source bitmap and puts
        their bounding boxes into a linked list.
*/
/*
extern "C" TCLIP_listNode * FindAllSprites(void *Source, int Border, int sizeX, int sizeY)
{
    TCLIP_listNode * RectList;
    TCLIP_listNode * NewNode;
    int count = 0;

    RectList = NULL;

    NewNode = FindSprite(Source, Border, sizeX, sizeY);
    NewNode->Prev = NULL;
    while (NewNode != NULL) {
        count++;
        NewNode->count = count;
        NewNode->Name           = NULL;

        NewNode->Flag           = 0;
        NewNode->Next           = RectList;
        RectList->Prev          = NewNode;
        RectList                = NewNode;

        NewNode = FindSprite(Source, Border, sizeX, sizeY);
    }

    return RectList;
}
*/

extern "C" TCLIP_listNode * CLIP_findAllSprites(void *Source, int Border, int sizeX, int sizeY,
                              void (*waitFunc)(long,long))
{

  TCLIP_listNode * RectList;
  TCLIP_listNode * NewNode;
  int count = 0;

  NewNode = RectList = CLIP_findSprite(Source, Border, sizeX, sizeY);
  if (NewNode)
    NewNode->Prev = NULL;
  while (RectList != NULL)
  {
    WAIT_func (waitFunc, count, 1000, 0, -1, -1);
    count++;
//    RectList->Next = NULL;
//    RectList->Prev = NULL;
    RectList->count = count;
    RectList->Name  = NULL;
    RectList->Flag  = 0;
    RectList->Next = CLIP_findSprite(Source, Border, sizeX, sizeY);
    if (RectList->Next)
    {
      RectList->Next->Prev = RectList;
      RectList->Next->Next = NULL;
    }
    RectList = RectList->Next;
  }
  WAIT_func (waitFunc, 1000, 1000, 0, -1, -1);

  return NewNode;
}


int CLIP_getRectsCount (TCLIP_listNode * Rects)
{
  int count = 0;

  while (Rects != NULL)
  {
    Rects = Rects->Next;
    count++;
  }

  return count;
}

/*

/*
        Function: PrintRects
    Description:
        Print all sprites found.
*/
/*
void PrintRects(TCLIP_listNode * Rects)
{
        printf("Found sprites at:\n");
        while (Rects != NULL) {
        printf("  (%d,%d) (%d,%d)\n",
                Rects->Rect.Left,
                        Rects->Rect.Top,
            Rects->Rect.Right,
            Rects->Rect.Bottom);
        Rects = Rects->Next;
    }
}
*/

/*
        Function: ExtractSprite
    Description:
        Creates a new sprite composed of the pixels enclosed by the
        rectangle passed to the function.  A linear bitmap of the
        enclosed are is passed back to the caller, or NULL if
        some sort of error occurs
*/

void  *CLIP_extractSprite (void *Source, int sizeX,
                      int Left, int Top, int Right, int Bottom)

{
    int Width;
    int Height;
    int WidthCounter;
    int HeightCounter;
    void *NewBitmap;
    unsigned char  * Data;

    Width = Right - Left - 1;
    Height = Bottom - Top - 1;

    NewBitmap = _alloc (Width * Height);
    if (NewBitmap)
    {
      /* -1 for the data byte in the struct itself */

  //    NewBitmap->Width    = Width;
  //    NewBitmap->Height   = Height;

      Data = (unsigned char  *) NewBitmap;
      for (HeightCounter = 0; HeightCounter < Height; HeightCounter++) {
          for (WidthCounter = 0; WidthCounter < Width; WidthCounter++) {
                  *Data++ = GetPixel(Source, Left + WidthCounter + 1,
                  Top + HeightCounter + 1, sizeX);
          }
      }
    }

    return NewBitmap;
}


TCLIP_listNode *CLIP_findSpriteAt (TCLIP_listNode *Rects, int x, int y)
{
  while (Rects != NULL)
  {
    if ((Rects->Rect.Top < y) && (Rects->Rect.Bottom > y) &&
        (Rects->Rect.Left < x) && (Rects->Rect.Right > x))
    {
      if (Rects->Flag == 0)
      {
        break;
      }
    }
    Rects=Rects->Next;
  }

  return Rects;
}
